""" Ground motion Kriger: Helpers
"""

import numpy as np

__author__ = 'A. Renmin Pretell Ductram'


def make_array(my_list):
	
	vals = []
	for val in zip(my_list):
		if isinstance(val[0],np.ndarray):
			vals.append(val[0])
		elif isinstance(val[0], (int, float)):
			vals.append(np.array([float(val[0])]))
		elif isinstance(val[0], (list, tuple)):
			vals.append(np.array(val[0], dtype=float))
		else:
			vals.append(val[0])
	return vals



""" Ground motion Kriger: Reader
"""

import os
import numpy as np
import json
from .core import Event
from .core import Fault
from .helpers import make_array

__author__ = 'A. Renmin Pretell Ductram'


def read_data(earthquake,gmim,model):
	
	with open(os.path.join(os.path.abspath(__file__).split('reader.py')[0],'Data','{}.json'.format(earthquake)), 'r') as f:
		data = json.load(f)
	
	stlat = np.array(data['WER']['lat_'+gmim.lower()])
	stlon = np.array(data['WER']['lon_'+gmim.lower()])
	stval = np.array(data['WER'][gmim])
	elat  = data['hypocenter']['hypo_lat']
	elon  = data['hypocenter']['hypo_lon']
	sill  = data['sill'][gmim]
	eta   = data['eta'][gmim]
	
	if model.lower() == 'realizations':
		Nmodel = 1000
		LE     = np.array(data['model']['all'][gmim]['le'])
		gammaE = np.array(data['model']['all'][gmim]['gammae'])
		LA     = np.array(data['model']['all'][gmim]['la'])
	
	elif model.lower() == 'map':
		Nmodel = 1
		LE     = np.array([data['model']['map'][gmim]['le']])
		gammaE = np.array([data['model']['map'][gmim]['gammae']])
		LA     = np.array([data['model']['map'][gmim]['la']])
	
	elif model.lower() == 'all':
		Nmodel = 1001
		LE     = np.array(data['model']['all'][gmim]['le'])
		LE     = np.append(LE,data['model']['map'][gmim]['le'])
		gammaE = np.array(data['model']['all'][gmim]['gammae'])
		gammaE = np.append(gammaE,data['model']['map'][gmim]['gammae'])
		LA     = np.array(data['model']['all'][gmim]['la'])
		LA     = np.append(LA,data['model']['map'][gmim]['la'])

	params = [*data['hypocenter'].values()]
	region = data['region']
	if region not in ['california','china','italy','japan','taiwan','turkey']:
		region = 'global'
	if earthquake in ['2023 M7.8 Pazarcik', '2023 M7.7 Kahramanmaras', '2023 M6.8 Nurdagi', '2023 M6.3 Yayladagi']:
		region = 'global'
	params = sum([[earthquake],[data['M']],params,[data['tectonic']], [region]],[])
	event = Event(*params)

	params = [*data['fault'].values()]
	params = make_array(params)
	fault  = Fault(*params)
	
	return stlat,stlon,stval,elat,elon,event,fault,sill,eta,Nmodel,LE,gammaE,LA



""" Ground motion Kriger: Main
"""

import os
import numpy as np
import pandas as pd
from .reader import read_data
from .core import Site
from .helpers import make_array
import geostats

__author__ = 'A. Renmin Pretell Ductram'


def get_Kgmim(ID,lat,lon,Vs30,earthquake,model,gmim):
	
	"""
	ID: Site name(s)
	lat: Site latitude(s)
	lon: Site longitude(s)
	Vs30: Site Vs30 value(s)
	Earthquake name: See list by using "gmKriger.models()"
	model: Corelation model, either "All", "Realizations", or "MAP"
	gmim: Ground motion intensity measure. See list by using "gmKriger.show()"
	"""
	
	pd.set_option('display.precision', 6)
	
	cwd  = os.path.abspath(__file__).split('gmKriger.py')[0]
	gmim = gmim.lower()
	
	# Check availability
	models = pd.read_csv(os.path.join(cwd,'Data','models.csv'))
	gmims  = models.columns[1:]
	
	if gmim.lower() not in gmims:
		print('Please double check the requested GMIM.')
		
	ind1 = earthquake in models['earthquake'].values
	ind2 = 0
	
	if ind1:
		ind2 = models[models['earthquake']==earthquake][gmim].values[0]
		
	if not ind1:
		print('No correlation models currently available for {}.'.format(earthquake))
		return
		
	elif ind1 and ind2 == 0:
		print('No {} correlation models currently available for {}'.format(gmim,earthquake))
		return
	
	else:
		
		# Start
		get_label1 = dict(zip(['pga','pgv','cav','ia','psa(0.100)','psa(0.300)','psa(0.600)','psa(1.000)','psa(3.000)','psa(6.000)','psa(10.000)','cavdp'],
							['PGA (g)','PGV (cm/s)','CAV (m/s)','Ia (m/s)','PSA(0.100) (g)','PSA(0.300) (g)','PSA(0.600) (g)','PSA(1.000) (g)','PSA(3.000) (g)','PSA(6.000) (g)','PSA(10.000) (g)','CAVdp (g-s)']))
		get_label2 = dict(zip(['pga','pgv','cav','ia','psa(0.100)','psa(0.300)','psa(0.600)','psa(1.000)','psa(3.000)','psa(6.000)','psa(10.000)','cavdp'],
							['PGA','PGV','CAV','Ia','PSA(0.100)','PSA(0.300)','PSA(0.600)','PSA(1.000)','PSA(3.000)','PSA(6.000)','PSA(10.000)','CAVdp']))
		label1 = get_label1[gmim]
		label2 = get_label2[gmim]
		
		# Check input
		if isinstance(ID, str):
			ID = [ID]
		[lat,lon,Vs30] = make_array([lat,lon,Vs30])
		if not len(ID)==len(lat)==len(lon)==len(Vs30):
			print('Unequal length of input.')
			return
		
		# Get data
		[stlat,stlon,stval,elat,elon,event,fault,sill,eta,Nmodel,LE,gammaE,LA] = read_data(earthquake,gmim,model)
		
		# Sampling a station
		sco  = pd.DataFrame(zip(lat,lon),columns=['lat','lon'])
		stco = pd.DataFrame(zip(stlat,stlon),columns=['lat','lon'])
		idx  = np.array(sco[sco.isin(stco.to_dict(orient='list')).all(axis=1)].index.tolist())
		
		# Kriging
		out_mu,out_va = [[],[]]
		for i in range(Nmodel):
			cmodel    = geostats.Model('ea',LE[i],gammaE[i],L_A=LA[i])
			sampled   = geostats.Site(stlat,stlon,value=stval)
			unsampled = geostats.Site(lat,lon)
			Kriger    = geostats.Kriging(sampled,unsampled,sill,cmodel,elat,elon)
			out       = Kriger.Krige('ordinary')
			out_mu.append(out.mean)
			out_va.append(out.variance)
		out_mu = np.array(out_mu)
		out_va = np.array(out_va)
		if len(idx) > 0:
			for i in range(Nmodel):
				out_va[i,:][idx] = 0
			
		# Ground motion
		fault.get_URC()
		fault.get_rfault()
		fault.get_fstyle()
		site = Site(lat, lon, Vs30, Z1p0=None, Z2p5=None)
		site.get_distances(fault)
		site.get_gmim(event, fault, gmim)

		if np.isnan(site.IM_mu[0]):
			return
		
		# Results
		if model.lower() in ['realizations','map']:
			K_WER_mean = np.mean(out_mu,axis=0)
			K_WER_std  = np.sqrt(np.mean(out_va,axis=0)+np.std(out_mu,axis=0)**2)
			K_IM_mu    = np.exp(site.IM_mu+[eta]*site.Nsites+K_WER_mean)
			df_out     = pd.DataFrame(zip(ID,lat,lon,K_IM_mu,K_WER_std),columns=['Site','Lat (deg)', 'Lon (deg)',label1,'sigma_{} (ln)'.format(label2)])

		else:
			K_WER_mean = np.mean(out_mu[:-1,:],axis=0)
			K_IM_mu    = np.exp(site.IM_mu+[eta]*site.Nsites+K_WER_mean)
			K_WER_std  = np.sqrt(np.mean(out_va[:-1,:],axis=0)+np.std(out_mu[:-1,:],axis=0)**2)
			if len(idx) > 0:
				K_WER_std[idx] = 0
			
			K_WER_mean_map = out_mu[-1,:]
			K_IM_mu_map    = np.exp(site.IM_mu+[eta]*site.Nsites+K_WER_mean_map)
			K_WER_std_map  = out_va[-1,:]**0.5
			
			df_out = pd.DataFrame(zip(ID,lat,lon,K_IM_mu_map,K_WER_std_map,K_IM_mu,K_WER_std),
				columns=['Site','Lat (deg)', 'Lon (deg)','{} - MAP'.format(label1),'sigma_{} (ln) - MAP'.format(label2),label1,'sigma_{} (ln)'.format(label2)])
		
		return df_out

def models():
	models = pd.read_csv(os.path.join(os.path.abspath(__file__).split('gmKriger.py')[0],'Data','models.csv'))
	models.replace(1, 'Yes', inplace=True)
	models.replace(0, 'No', inplace=True)
	return models



""" Ground motion Kriger: Core
"""

import numpy as np
import gmms
from gmms import get_fault_URC
from gmms import get_representative_fault
from gmms import get_faulting_style
from gmms import get_distances_cy
from .helpers import make_array
import pygmm

__author__ = 'A. Renmin Pretell Ductram'


class Site:
	
	def __init__(self, latitude, longitude, Vs30=None, Z1p0=None, Z2p5=None):
		self.Nsites    = len(latitude)
		self.latitude  = latitude
		self.longitude = longitude
		self.IM_mu  = np.full(self.Nsites, np.nan)
		self.IM_phi = np.full(self.Nsites, np.nan)
		self.IM_tau = np.full(self.Nsites, np.nan)

		if Vs30 is not None:
			self.Vs30 = Vs30

		if Z1p0 is None and Vs30 is not None:
			self.Z1p0 = np.exp((-7.15/4)*np.log((self.Vs30**4+571**4)/(1360**4+571**4))) # BSSA14
		else:
			self.Z1p0 = Z1p0

		if Z2p5 is None and Vs30 is not None:
			self.Z2p5 = np.exp(7.089-1.144*np.log(self.Vs30)) # CB13
		else:
			self.Z2p5 = Z2p5
	
	def get_distances(self,fault):
		params = [*fault]
		params = sum([[self.latitude.tolist(),self.longitude.tolist()],params],[])
		params = make_array(params)
		self.Rjb, self.Rrup, self.Rx = get_distances_cy(*params)

	def get_gmim(self, event, fault, gmim, flag_CAVgm=None, flag_PSV=None):
		self.gmim   = gmim
		if flag_CAVgm is None:
			self.flag_CAVgm = 1
		if flag_PSV is None:
			self.flag_PSV = 1
		
		if gmim.lower() == 'pga':
			self.gmim_tag = 0
		elif gmim.lower() == 'pgv':
			self.gmim_tag = 1
		elif gmim.lower() == 'ia':
			self.gmim_tag = 2
		elif gmim.lower() == 'cav':
			self.gmim_tag = 3
		elif gmim.lower() == 'cavdp':
			self.gmim_tag = 4
		elif gmim.lower() == 'psa(0.100)':
			self.gmim_tag = 5
		elif gmim.lower() == 'psa(0.300)':
			self.gmim_tag = 6
		elif gmim.lower() == 'psa(0.600)':
			self.gmim_tag = 7
		elif gmim.lower() == 'psa(1.000)':
			self.gmim_tag = 8
		elif gmim.lower() == 'psa(3.000)':
			self.gmim_tag = 9
		elif gmim.lower() == 'psa(6.000)':
			self.gmim_tag = 10
		elif gmim.lower() == 'psa(10.000)':
			self.gmim_tag = 11

		IM_mu  = np.zeros(self.Nsites)
		IM_phi = np.zeros(self.Nsites)
		IM_tau = np.zeros(self.Nsites)
		
		if self.gmim_tag == 0:
			if event.tectonic.lower() == 'crustal':
				for i in range(self.Nsites):
					scenario = pygmm.Scenario(mag=event.M,depth_1_0=self.Z1p0[i],dist_jb=self.Rjb[i],v_s30=self.Vs30[i],mechanism=fault.fmech,region=event.region)
					gmm_estimate = pygmm.BooreStewartSeyhanAtkinson2014(scenario)
					IM_mu[i]  = np.log(gmm_estimate.pga)
					IM_phi[i] = gmm_estimate._phi[0]
					IM_tau[i] = gmm_estimate._tau[0]
			else:
				print('No ground motion models currently implemented for {} events.'.format(event.tectonic))
				return

		elif self.gmim_tag == 1:
			if event.tectonic.lower() == 'crustal':
				for i in range(self.Nsites):
					scenario = pygmm.Scenario(mag=event.M,depth_1_0=self.Z1p0[i],dist_jb=self.Rjb[i],v_s30=self.Vs30[i],mechanism=fault.fmech,region=event.region)
					gmm_estimate = pygmm.BooreStewartSeyhanAtkinson2014(scenario)
					IM_mu[i]  = np.log(gmm_estimate.pgv)
					IM_phi[i] = gmm_estimate._phi[-1]
					IM_tau[i] = gmm_estimate._tau[-1]
			else:
				print('No ground motion models currently implemented for {} events.'.format(event.tectonic))
				return
		
		elif self.gmim_tag in [5,6,7,8,9,10,11]:
			T = float(gmim[4:9])
			periods = pygmm.model.load_data_file('boore_stewart_seyhan_atkinson-2014.csv', 2).period
			boo_IM  = periods==T
				
			for i in range(self.Nsites):
				scenario = pygmm.Scenario(mag=event.M,depth_1_0=self.Z1p0[i],dist_jb=self.Rjb[i],v_s30=self.Vs30[i],mechanism=fault.fmech,region=event.region)
				gmm_estimate = pygmm.BooreStewartSeyhanAtkinson2014(scenario)
				IM_mu[i]  = np.log(gmm_estimate.spec_accels[boo_IM[2:]][0])
				IM_phi[i] = gmm_estimate._phi[boo_IM][0]
				IM_tau[i] = gmm_estimate._tau[boo_IM][0]

		elif self.gmim_tag == 2:
			if event.tectonic.lower() == 'crustal' and event.region.lower() != 'japan':
				[IM_mu,IM_phi,IM_tau] = gmms.CampbellBozorgnia2019.CampbellBozorgnia2019_cy(0,event.M,fault.rwidth,fault.rdip,fault.rZtor,event.hyp_Z,self.Rjb,self.Rrup,self.Rx,self.Vs30,self.Z2p5,fault.fnm,fault.frv,event.region)
			else:
				[IM_mu,IM_phi,IM_tau] = gmms.FoulserPiggottGoda2015.FoulserPiggottGoda2015_cy(0,event.M,event.hyp_lat,event.hyp_lon,event.hyp_Z,self.Rrup,self.Vs30,event.fins,event.fint,fault.fnm,fault.frv)

		elif self.gmim_tag == 3:
			if event.tectonic.lower() == 'crustal' and event.region.lower() != 'japan':
				[IM_mu,IM_phi,IM_tau] = gmms.CampbellBozorgnia2019.CampbellBozorgnia2019_cy(1,event.M,fault.rwidth,fault.rdip,fault.rZtor,event.hyp_Z,self.Rjb,self.Rrup,self.Rx,self.Vs30,self.Z2p5,fault.fnm,fault.frv,event.region)
			else:
				[IM_mu,IM_phi,IM_tau] = gmms.FoulserPiggottGoda2015.FoulserPiggottGoda2015_cy(1,event.M,event.hyp_lat,event.hyp_lon,event.hyp_Z,self.Rrup,self.Vs30,event.fins,event.fint,fault.fnm,fault.frv)

		elif self.gmim_tag == 4:
			if event.tectonic.lower() == 'crustal':
				[IM_mu,IM_phi,IM_tau] = gmms.CampbellBozorgnia2011.CampbellBozorgnia2011_cy(event.M,fault.rdip,fault.rZtor,self.Rjb,self.Rrup,self.Rx,self.Vs30,self.Z1p0,self.Z2p5,fault.fnm,fault.frv,self.flag_CAVgm,self.flag_PSV)
			else:
				print('No ground motion models currently implemented for {} events.'.format(event.tectonic))
				return

		self.IM_mu  = IM_mu
		self.IM_phi = IM_phi
		self.IM_tau = IM_tau


class Event:

	def __init__(self,name,M, hyp_lat, hyp_lon, hyp_Z, tectonic, region):
		self.name     = name
		self.M        = M
		self.hyp_lat  = hyp_lat
		self.hyp_lon  = hyp_lon
		self.hyp_Z    = hyp_Z
		self.tectonic = tectonic
		self.region   = region

		if self.tectonic.lower() == 'interface':
			self.fins = 0
			self.fint = 1
		elif self.tectonic.lower() == 'intraslab':
			self.fins = 1
			self.fint = 0
		else:
			self.fins = 0
			self.fint = 0


class Fault:

	def __init__(self, length, width, strike, dip, rake, ULC_lat, ULC_lon, Ztor, flag=None):
		self.Nsegm   = len(length)
		self.length  = length
		self.width   = width
		self.strike  = strike
		self.dip     = dip
		self.rake    = rake
		self.ULC_lat = ULC_lat
		self.ULC_lon = ULC_lon
		self.Ztor    = Ztor

	def __iter__(self):
		yield self.ULC_lat.tolist()
		yield self.ULC_lon.tolist()
		yield self.URC_lat.tolist()
		yield self.URC_lon.tolist()
		yield self.width.tolist()
		yield self.length.tolist()
		yield self.dip.tolist()
		yield self.strike.tolist()
		yield self.Ztor.tolist()

	def get_URC(self):
		self.URC_lat, self.URC_lon = get_fault_URC(self.ULC_lat, self.ULC_lon, self.length, self.strike)

	def get_rfault(self):
		self.rrake,self.rwidth,self.rdip,self.rZtor = get_representative_fault(self.rake,self.width,self.length,self.dip,self.Ztor)

	def get_fstyle(self):
		self.fnm,self.frv,self.fmech = get_faulting_style(self.rrake)
